
class ModelingObjectBase(object):
    """ Base class for objects belonging to a model
        INTERNAL class.
    """

    # @profile
    def __init__(self, model, name=None):
        self._name = name if name is None else str(name)
        self.__model = model

    def get_name(self):
        return self._name

    def _set_name(self, name):
        self.check_name(name)
        self._name = name

    def check_name(self, new_name):
        """
        Redefine this method to check a new candidate name.
        """
        pass

    name = property(get_name, _set_name)

    def has_name(self):
        return self.name is not None

    @property
    def model(self):
        return self.__model

    def is_in_model(self, model):
        return model and self.__model == model

    def _check_model_has_solution(self):
        self.model.check_has_solution()

    def fatal(self, msg, *args):
        self.error_handler.fatal(msg, args)

    def error(self, msg, *args):
        self.error_handler.error(msg, args)

    def warning(self, msg, *args):
        self.error_handler.warning(msg, args)

    def debug(self, msg, *args):
        self.error_handler.trace(msg, args)

    @property
    def error_handler(self):
        return self.__model.error_handler

    def name_or_str(self):
        """
        :return: A string which is either the name or the literal expression of the object.
        """
        return self.name or str(self)


class ModelingObject(ModelingObjectBase):
    """ Base class for all modeling objects that have an engine id (index).
    """

    @staticmethod
    def is_valid_index(idx):
        # INTERNAL: This is where the valid index check is performed
        return idx >= 0

    # @profile
    def __init__(self, model, name=None, index=-1, is_automatic_name=False):
        ModelingObjectBase.__init__(self, model, name)
        self.__index = index
        self._has_automatic_name = is_automatic_name

    def has_automatic_name(self):
        return self._has_automatic_name

    def has_user_name(self):
        """ Checks whether the object has a valid name given by the user.

        """
        return self.has_name() and not self._has_automatic_name

    def notify_automatic_name(self):
        self._has_automatic_name = True

    def is_generated(self):
        """ Checks whether this object has been generated by another modeling object.

        If so, the origin object is stored in the "_origin" attribute.

        Returns:
            True if the objects has been generated.
        """
        try:
            is_generated = self._origin is not None
            return is_generated

        except AttributeError:
            return False

    def notify_origin(self, origin):
        if origin is not None:
            self._origin = origin

    def origin(self):
        return self._origin if hasattr(self, "_origin") else None

    def __hash__(self):
        return id(self)

    @property
    def unchecked_index(self):
        return self.__index

    def get_index(self):
        return self.__index

    def set_index(self, idx):
        self.__index = idx

    def has_valid_index(self):
        return ModelingObject.is_valid_index(self.__index)

    def _get_index_or_raise(self):
        ''' Returns the index if valid, otherwise raises an exception.'''
        if not self.has_valid_index():
            self.fatal("Modeling object {0!s} has invalid index: {1:d}", self, self.__index)  # pragma : no cover
        return self.__index

    index = property(get_index, set_index)
    safe_index = property(_get_index_or_raise)

    def to_linear_expr(self):
        """Redefine this method for those objects which can be mapped to
        linear expressions.
        """
        self.fatal("Not implemented: to_linear_exp")  # pragma : no cover


class Expr(ModelingObjectBase):
    """ Parent class for all expression classes.
    """

    def __init__(self, model, name=None):
        ModelingObjectBase.__init__(self, model, name)

    def clone(self):
        raise NotImplementedError  # pragma: no cover

    def copy(self, target_model, var_mapping):
        # internal
        raise NotImplementedError  # pragma : no cover

    def number_of_variables(self):
        """ Returns the number of variables in the expression.
        """
        return sum(1 for _ in self.iter_variables())  # pragma : no cover

    def iter_variables(self):
        """ Returns an iterator over all variables mentioned in the expression.
        """
        raise NotImplementedError  # pragma: no cover

    def contains_var(self, dvar):
        """ Checks whether a variable is mentioned in the expression.

        Args:
            dvar: An instance of Var.
        """
        raise NotImplementedError  # pragma: no cover

    def get_coef(self, avar):
        """ Returns the coefficient of a variable in the expression.

        Note
            If the variable is not present in the expression, the function returns 0.

        :param avar: The variable for which the coefficient is being queried.

        :return: A floating-point value.
        """
        self.model.typecheck_var(avar)
        return self.unchecked_get_coef(avar)

    def unchecked_get_coef(self, avar):
        raise NotImplementedError  # pragma : no cover

    def __contains__(self, avar):
        """Overloads operator "in" for an expression and a variable.

        Args:
            avar: A decision variable.

        Returns:
            True if ``contains_var`` is true.
        """
        return self.contains_var(avar)

    def _num_to_stringio(self, oss, num, fmt):
        # INTERNAL
        if num == int(num):
            oss.write('%d' % num)
        else:
            oss.write(fmt % num)

    def _num_to_stringio2(self, oss, num, ndigits):
        from docplex.utils import is_int
        if is_int(num):
            oss.write('%d' % num)
        else:
            # use second arg as nb digits:
            oss.write("{0:.{1}f".format(num, ndigits))

    def __pos__(self):
        # + e is identical to e
        return self