
from docplex.mp.solution import SolveSolution
from docplex.mp.utils import DOCplexException
from docloud.status import JobSolveStatus
# gendoc: ignore


class ISolver(object):
    """
    The pure solving part
    """
    def can_solve(self):
        """
        :return: True if this engine class can truly solve
        """
        raise NotImplementedError  # pragma: no cover

    def connect_progress_listeners(self, listeners):
        """
        Connects progress listeners
        :param listeners:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def solve(self, mdl):
        ''' Redefine this method for the real solve.
            Returns True if model is well-formed and a solution has been found.
        '''
        raise NotImplementedError  # pragma: no cover

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        """
        Runs feasopt-like algorithm with a set of relaxable cts with preferences
        :param relaxable_groups:
        :param optimize:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def get_solve_status(self):
        """  Return a DOcloud-style solve status.

        Possible enums are in docloud/status.py
        Default is UNKNOWN at this stage. Redefined for CPLEX and DOcloud engines.
        """
        return JobSolveStatus.UNKNOWN

    def get_solutions(self, dvars):
        """
        Returns a dictionary of variable solution values.
        :param dvars:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def get_cplex(self):
        """
        Returns the underlying CPLEX, if any. May raise an exception if not applicable.
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def has_cplex(self):
        try:
            return self.get_cplex() is not None
        except DOCplexException:
            # some engine may raise an exception when accessing a cplex
            return False

    def set_parameter(self, parameter, value):
        """ Changes the parameter value.
        :param parameter:
        :param value:
        """
        raise NotImplementedError

    def get_parameter(self, parameter):
        raise NotImplementedError

    def is_parameter_nondefault(self, parameter):
        return self.get_parameter(parameter) != parameter.default_value


# noinspection PyAbstractClass
class IEngine(ISolver):
    """ interface for all engine facades
    """

    def name(self):
        ''' Returns the code to be used in model'''
        raise NotImplementedError  # pragma: no cover

    def get_var_index(self, var):
        raise NotImplementedError  # pragma : no cover

    def get_ct_index(self, index):
        raise NotImplementedError  # pragma : no cover

    def getInfinity(self):
        raise NotImplementedError  # pragma: no cover

    def createOneVariable(self, vartype, lb, ub, name):
        raise NotImplementedError  # pragma: no cover

    def createVariables(self, keys, vartype, lb, ub, namer):
        raise NotImplementedError  # pragma: no cover

    def createBinaryLinearConstraint(self, binaryct):
        raise NotImplementedError  # pragma: no cover

    def createRangeConstraint(self, rangect):
        raise NotImplementedError  # pragma: no cover

    def createIndicatorConstraint(self, ind):
        raise NotImplementedError  # pragma: no cover

    def remove_constraint(self, ct):
        raise NotImplementedError  # pragma: no cover

    def set_objective(self, sense, expr):
        raise NotImplementedError  # pragma: no cover

    def clear_objective(self, expr):
        raise NotImplementedError  # pragma: no cover

    def end(self):
        raise NotImplementedError  # pragma: no cover

    def notify_trace_mode(self, arg):
        raise NotImplementedError  # pragma: no cover

    def set_var_attribute(self, var, attr_name, attr_val):
        raise NotImplementedError  # pragma : no cover

    def get_var_attribute(self, var, attr_name):
        raise NotImplementedError  # pragma : no cover


# noinspection PyAbstractClass
class DummyEngine(IEngine):

    def createRangeConstraint(self, rangect):
        return -1

    def createIndicatorConstraint(self, ind):
        return -1

    def notify_trace_mode(self, arg):
        pass

    def getInfinity(self):
        return 1e+20

    def get_var_index(self, var):
        return -1

    def get_ct_index(self, index):
        return -1

    def createOneVariable(self, vartype, lb, ub, name):
        return -1

    def createVariables(self, keys, vartype, lb, ub, namer):
        return [-1] * len(keys)

    def set_var_attribute(self, avar, attr_name, attr_val):
        return attr_val

    def get_var_attribute(self, var, attr_name):
        if "name" == attr_name:
            return var.name
        elif "lb" == attr_name:
            return var.lb
        elif "ub" == attr_name:
            return var.ub
        else:
            raise NotImplementedError

    def createBinaryLinearConstraint(self, binaryct):
        return -1

    def remove_constraint(self, ct):
        pass

    def set_objective(self, sense, expr):
        pass

    def end(self):
        """ terminate the engine
        """
        pass

    def connect_progress_listeners(self, listeners):
        pass

    def can_solve(self):
        return False

    def solve(self, mdl):
        return False

    def get_solve_status(self):
        return JobSolveStatus.UNKNOWN

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        raise NotImplementedError  # pragma : no cover

    def get_solve_attribute(self, attr_name, indices):
        return {}

    def get_solutions(self, *args):
        return {}

    def clear_objective(self, expr):
        pass

    def get_cplex(self):
        raise DOCplexException("No CPLEX is available.")


# noinspection PyAbstractClass
class IndexerEngine(DummyEngine):
    """
    An abstract engine facade which generates unique indices for variables, constraints
    """

    def __init__(self, initial_index=0):
        DummyEngine.__init__(self)
        self._initial_index = initial_index  # CPLEX indices start at 0, not 1
        self.__var_counter = self._initial_index
        self.__ct_counter = self._initial_index
        # a local dictionary cache for parameters.
        self._local_params = {}

    def _incrementVars(self, size=1):
        self.__var_counter += size
        return self.__var_counter

    def _incrementCts(self, size=1):
        self.__ct_counter += size
        return self.__ct_counter

    def createOneVariable(self, vartype, lb, ub, name):
        old_count = self.__var_counter
        self._incrementVars(1)
        return old_count

    def createVariables(self, keys, vartype, lb, ub, namer):
        oldCount = self.__var_counter
        newCount = self._incrementVars(len(keys))
        return list(range(oldCount, newCount))

    def _create_one_ct(self):
        old_ct_count = self.__ct_counter
        self._incrementCts(1)
        return old_ct_count

    def createBinaryLinearConstraint(self, binaryct):
        return self._create_one_ct()

    def createRangeConstraint(self, rangect):
        return self._create_one_ct()

    def createIndicatorConstraint(self, ind):
        return self._create_one_ct()

    def get_solve_attributes(self, attr_name, indices):
        # return empty dict
        return {}

    def dump(self, path):
        pass

    def set_objective(self, sense, expr):
        pass

    def clear_objective(self, expr):
        pass


    def set_parameter(self, parameter, value):
        """ Changes the parameter value.

        For this limited type of engine, it stores the value in a local dictionary.
        Note:
            This method assumes the caller method has already checked whether the value
            was acceptable. It only deals with the engine side.

        :param parameter:
        :param value:
        """
        self._local_params[parameter] = value

    def get_parameter(self, parameter):
        """ Gets the current value of a parameter.

        If the parameter has been set, go to the local cache,
        otherwise return the parameter's default value.

        Params:
         parameter: the parameter for which we query the value.

        """
        return self._local_params.get(parameter, parameter.default_value)


    def is_parameter_nondefault(self, parameter):
        if parameter not in self._local_params:
            return False
        else:
            return self._local_params[parameter] != parameter.default_value


class LocalEngine(IndexerEngine):
    """
    """
    def __init__(self, error_handler, docloud_context=None):
        IndexerEngine.__init__(self)

    def name(self):
        return "local"

    def get_var_index(self, var):
        return var.index

    def get_ct_index(self, ct):
        return ct.index

    def can_solve(self):
        return False

    def solve(self, mdl):
        """
        This solver cannot solve. never ever.
        """
        mdl.fatal("No CPLEX DLL and no DOcloud credentials: model cannot be solved!")
        return None

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        return False, 0


class FakeSuccessEngine(IndexerEngine):
    """
    A fake engine which returns some silly objective, used for performance testing
    """
    def __init__(self, error_handler, docloud_context=None):
        IndexerEngine.__init__(self)

    def name(self):
        return "yes"

    def solve(self, mdl):
        # sets all variable values to zero
        return SolveSolution(mdl, obj=0, var_value_map=None, engine_name=self.name, keep_zeros=False)

    def can_solve(self):
        return True

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        return True, 0


class FakeFailEngine(IndexerEngine):
    """
    A fake engine which returns some silly objective, used for performance testing
    """
    def __init__(self, error_handler, docloud_context=None):
        IndexerEngine.__init__(self)

    def name(self):
        return "fail"

    def solve(self, mdl):
        # always fail
        return None

    def can_solve(self):
        return True

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        return False, 0

    def get_solve_status(self):
        return JobSolveStatus.INFEASIBLE_SOLUTION