

class KPI(object):
    """ Abstract class for KPIs.

    KPIs have a unique name.

    KPIs are attached to a model instance and can compute a numerical value,
    using the compute() method.

    Some KPIs require a valid solution of the model, while others do not. You can check whether a
    given KPI requires a solution by using requires_solution()

    """

    def __init__(self):
        pass

    def get_name(self):
        """ Returns the published name of the KPI.

        The published name is a non-empty, unique string identifying the KPI.

        """
        raise NotImplementedError

    def get_model(self):
        """  Returns the model instance on which the KPI is defined.
        """
        raise NotImplementedError

    def compute(self):
        raise NotImplementedError

    def _check_name(self, name_arg):
        if not isinstance(name_arg, str):
            self.get_model().fatal("KPI_name, not a string: {0!s}", (name_arg,))
        elif not name_arg:
            self.get_model().fatal("KPI_name must be non-empty string, got: {0!s}", (name_arg,))
        else:
            pass

    def requires_solution(self):
        """ Returns True if the KPI requires a valid solution.

        KPIs based on decision expressions or variables require a successful solution
        to be computed. KPIs based on functions may not require a solution.

        """
        raise NotImplementedError

    def copy(self, new_model, var_map):
        raise NotImplementedError


class DecisionKPI(KPI):

    # Implementation of KPI abstract class
    # with a decision object.
    # Decision objects have a "solution_value" property.
    # This kind of KPI can only be computed after a solve has been performed.

    def __init__(self, decision_obj, name=None):
        KPI.__init__(self)
        self._dobj = decision_obj.to_linear_expr()
        self._name = name or decision_obj.name
        self._check_name(self._name)

    def get_name(self):
        return self._name

    name = property(get_name)

    def get_model(self):
        return self._dobj.model

    def compute(self):
        return self._dobj.solution_value

    def requires_solution(self):
        return True

    def get_expression(self):
        return self._dobj

    def copy(self, new_model, var_map):
        dobj_copy = self._dobj.copy(new_model, var_map)
        return DecisionKPI(decision_obj=dobj_copy, name=self.name)


class FunctionalKPI(KPI):
    # Functional KPIs store a function that takes a model to compute a number
    # Functional KPIs do not require a successful solve.

    def __init__(self, fn, model, name):
        KPI.__init__(self)
        self._name = name
        self._function = fn
        self._model = model
        self._check_name(self._name)

    def get_name(self):
        return self._name

    name = property(get_name)

    def get_model(self):
        return self._model

    def compute(self):
        return self._function(self._model)

    def requires_solution(self):
        return False

    def copy(self, new_model, var_map):
        return FunctionalKPI(fn=self._function, model=self._model, name=self._name)
